"""This module provides the base class for drugref API database adapters
All drugref adapters should inherit from this base class
(c)2005 by Dr Horst Herb <hherb@drugref.org>
This is free software released under the GPL license version 2 or newer
Details of the GPL license condtion can be obtained at http://www.gnu.org/licenses/gpl.txt
"""

class NotImplementedError:
	"""This error is thrown whenever somebody tries to access a API method that has not been
	implemented for that particular database"""
	pass

class drugrefapi:
	"""Base class for all drugref API backend adapters"""
	def __init__(self, host='localhost', port=8123, database='drugref', user='drugref', pwd='drugref', backend='postgres'):
		"""The vast majority of adapters will probably use a SQL server backend. Adapters that don't need such
		connection parameters, can simply ignore them."""
		self._name = "generic adapter"
		self._version = '0.2'
		self._versiondate = '11.01.2005'
		self._params={}
		self._params['host']=host
		self._params['port']=port
		self._params['database']=database
		self._params['user']=user
		self._params['pwd']=pwd
		self._params['backend']=backend
		self._dbapi = None
		self._db = None
		self._wildcard = "*"
		self._languages = ['?']
		self._countries = ['?']
		if backend == 'postgres':
			self.wildcard = "%"
			self._notcasesensitive = 'ilike'
		self._queries = {}	#available queries; key is the query name, value is a tuple
					#containing  the actual query string, and the required search expression
					#variables as list of key:value pairs with keys being the variable names,  
		self._provided_by = {}
		self._cache = {}	#cached query results. 
					#key is the provided data item, values is a tuple containing it's value
					#and the primary key used to retrieve it
		self._provided = {}	#data items that are provided by this drugref adapter
					#key is data item, value is the query name that can retrieve that item
		self._required = {}	#search expressions and primary keys that produced cached query results; 
					#key is the query name, value is the search expression or primary key
		self._funcs = {}	
		
	def addquery(self, query, requires, provides):
		"""register a query that will allow to find and retrieve attributes.
		query: the query string, with parameter placeholdrs as per Python DBAPI 2.0
		requires: symbolic search expression keyword, as described in the drugref user manual
		provides: list of attribute names as described in the drugref user manual
		RETURNS: nothing""" 
		for provided in provides:
			#print "registering attribute", provided, "] for ", self._name
			self._provided[provided]={'query' : query, 'requires' : requires}
		self._provided_by[query] = provides 
			
	def addfunc(self,func, requires, provides):
		"""register a function that will allow to find and retrieve attributes.
		func: a function that accepts a "key" as parameter and returns a list of lists of key:value pairs
		requires: symbolic search expression keyword, as described in the drugref user manual
		provides: list of attribute names as described in the drugref user manual
		RETURNS: nothing"""
		for provided in provides:
			print "registering attribute [", provided, "] for ", func, "in", self._name
			self._provided[provided]={'func' : self._funcs[func], 'requires' : requires} 
		self._provided_by[func] = provides
		
		
	def find(self, attribute, key):
		"""Search for a drugref attribute as identified by parameter "key"
		Possible attributes and keys have to be looked up in the drugref
		user manual.
		Searches are case insensitive, wildcars are allowed (see self.get_wildcard() )
		RETURNS: a list of key:value pairs. Possible keys have to be looked
		up in the drugref user manual.  
		"""
		#check whether we can provide this attribute in principle (= service provider available)
		if not self._provided.has_key(attribute):
			return {'Error':'Attribute "%s" not available' % attribute}
		
		queryinfo = self._provided[attribute]
		try:
			#if service provider available, try first to retrieve the attribute by registered function
			results = queryinfo['func'](key)
		        #else check whether we can retrieve the attribute via registered query
		except KeyError:
			#Function not available for attribute, trying query ...
			try:
				self._connect()
				try:
					query = self._queries[queryinfo['query']]
				except KeyError:
					#No query available for the requested attribute - this should never happen
					print "Unexpected Error - no query available for attribute", attribute 
					return [{'Error' : 'Query not available for attribute'}]
				params = {queryinfo['requires'] : key}
				#print query, params
				results = self._runquery(query, params)
			except KeyError:
				results=None
		if not results:
			return [{'Error' : 'Not Found'}]
		return results
		
			
	def get(self, attribute, key):
		"""Access a drugref attribute as identified by parameter "key"
		Possible attributes and keys have to be looked up in the drugref
		user manual.
		RETURNS: attribute value if available, else empty string
		"""
		datasource=None #flag whether a functin or a query was used to retrieve the attribute
		#check whether we can provide this attribute in principle (= service provider available)
		if not self._provided.has_key(attribute):
			return ''
		
		#print "Trying to access attribute", attribute, "via key = [%s]" % key
		try:
			#check whether we have this attribute in cache
			attr = self._cache[attribute]
			#check whether the cached attribute is assciated with the identifying key
			if attr[1]== key:
				return attr[0]
		except:
			pass
		
		queryinfo = self._provided[attribute]
		try:
			#if service provider available, try first to retrieve the attribute by registered function
			func = queryinfo['func']
			results = func(key)
			datasource = func
                #else check whether we can retrieve the attribute via registered query
		except KeyError:
			#Function not available for attribute, trying query ...
			self._connect()
			try:
				queryname = queryinfo['query']
			except:
				#No query available for the requested attribute - this should never happen
				print "Unexpected Error - no query available for attribute", attribute 
				return ''
			query = self._queries[queryname]
			datasource = queryname
			params = {queryinfo['requires'] : key}
			results = self._runquery(query, params)
		if not results:
			return ''
		#let's fill the cache with the attributes
		#it should have been a primary key search, so there should only be one result
		assert(len(results)==1)
		result = results[0]
		for attr in self._provided_by[datasource]:
			self._cache[attr] = (result[attr], key) 
		return self._cache[attribute][0]
		
	
	def _connect(self):
		"""connect to the database backend.
		RETURNS: nothing"""
		if self._params['backend'] == 'postgres':
			if not self._dbapi:
				import pyPgSQL.PgSQL as dbapi
				self._dbapi = dbapi
				self._dbapi.fetchReturnsList=0
			if not self._db:
				self._db = self._dbapi.connect(database = self._params['database'])
				
			
	def _runquery(self, query, params=None):
		""" run the query "query" on a dbapi 2.0 compliant backend adapter,
		optionally using the query parameters "params".
		RETURNS: list of result sets"""
		self._connect()
		cur = self._db.cursor()
		if params:
			cur.execute(query, params)
		else:
			cur.execute(query)
		result = cur.fetchall()
		return result
			
	def get_wildcard(self):
		"""get the wildcard that can be used in queries to represent any number of any characters
		RETURNS: character"""
		return self._wildcard

	def get_name(self):
		"""identify ourselves
		RETURNS: string"""
		return self._name
		
	def get_version(self):
		return self._version
				
	def list_capabilities(self):
		"""list all attributes we can provide
		RETURNS: list of strings"""
		return self._provided
		